/*
 * SIP Communicator, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package net.java.sip.communicator.impl.gui.keybindings;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;

import net.java.sip.communicator.impl.gui.keybindings.chooser.BindingAdaptor;
import net.java.sip.communicator.impl.gui.keybindings.chooser.BindingChooser;
import net.java.sip.communicator.impl.gui.keybindings.chooser.BindingEntry;
import net.java.sip.communicator.service.gui.ConfigurationForm;
import net.java.sip.communicator.service.gui.keybindings.KeybindingSet;
import net.java.sip.communicator.service.gui.keybindings.KeybindingsService;

/**
 * The <tt>ConfigurationForm</tt> that would be added to the settings
 * configuration to configure the application keybindings.
 * @author Damian Johnson
 */
public class KeybindingsConfigForm
    extends JPanel
    implements ConfigurationForm
{
    private static final long serialVersionUID = 0;
    private HashMap <KeybindingSetImpl, SIPChooser> choosers =
            new HashMap <KeybindingSetImpl, SIPChooser>();
    
    public KeybindingsConfigForm(KeybindingsService service)
    {
        super(new BorderLayout());
        
        setFocusable(true);
        JTabbedPane chooserPanes = new JTabbedPane(JTabbedPane.LEFT);
        
        // deselects entries awaiting input when focus is lost
        this.addFocusListener(new FocusAdapter()
        {
            public void focusLost(FocusEvent event)
            {
                for (SIPChooser chooser : choosers.values())
                {
                    chooser.setSelected(null, false);
                }
            }
        });
        
        for (KeybindingSet.Category category : KeybindingSet.Category.values())
        {
            KeybindingSetImpl bindingSet =
                    (KeybindingSetImpl) service.getBindings(category);
            if (bindingSet == null) continue; // defaults failed to load
            
            SIPChooser newChooser = new SIPChooser();
            for (BindingWrapper wrapper : bindingSet.getWrappedBindings())
            {
                newChooser.putBinding(wrapper.shortcut, wrapper.altShortcut,
                        wrapper.action);
            }
            
            // adds labels to top
            JPanel chooserWrapper = new JPanel(new BorderLayout());
            chooserWrapper.add(newChooser
                    .makeLabels(BindingChooser.FrameColorScheme.GRAY_DERIVED),
                    BorderLayout.NORTH);
            chooserWrapper.add(newChooser);
            
            // makes content aligned from top
            JPanel topWrapper = new JPanel(new BorderLayout());
            topWrapper.add(chooserWrapper, BorderLayout.NORTH);
            JScrollPane scroller = new JScrollPane(topWrapper);
            
            // adds listener that receives events to set bindings
            BindingAdaptor shortcutSetter = newChooser.makeAdaptor();
            
            // allows alternates to be removed
            shortcutSetter.setAlternateBindingsDisableable(true);
            this.addKeyListener(shortcutSetter);
            
            chooserPanes.addTab(getReadableConstant(category.toString()),
                    scroller);
            this.choosers.put(bindingSet, newChooser);
        }
        
        add(chooserPanes);
        
        // reverts bindings to defaults (un-applied)
        JButton defaults = new JButton(Resources.getString("useDefaults"));
        defaults.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                for (KeybindingSetImpl set : choosers.keySet())
                {
                    SIPChooser chooser = choosers.get(set);
                    chooser.clearBindings();
                    for (BindingWrapper wrapper : set.getDefaultBindings())
                    {
                        chooser.putBinding(wrapper.shortcut,
                                wrapper.altShortcut, wrapper.action);
                    }
                }
            }
        });
        
        JButton apply = new JButton(Resources.getString("apply"));
        apply.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                for (KeybindingSetImpl set : choosers.keySet())
                {
                    set.setBindings(choosers.get(set).getWrappedBindings());
                }
            }
        });
        
        JPanel bottomWrapper = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        bottomWrapper.add(defaults);
        bottomWrapper.add(apply);
        add(bottomWrapper, BorderLayout.SOUTH);
    }
    
    /**
     * Implements the <tt>ConfigurationForm.getTitle()</tt> method. Returns
     * the title of this configuration form.
     */
    public String getTitle()
    {
        return Resources.getString("keybindings");
    }
    
    /**
     * Implements the <tt>ConfigurationForm.getIcon()</tt> method. Returns the
     * icon of this configuration form.
     */
    public byte[] getIcon()
    {
        return Resources.getImageInBytes("keybindingPluginIcon");
    }
    
    /**
     * Implements the <tt>ConfigurationForm.getForm()</tt> method. Returns the
     * component corresponding to this configuration form.
     */
    public Object getForm()
    {
        return this;
    }
    
    /**
     * Provides a more readable version of constant names. Spaces replace
     * underscores and this changes the input to lowercase except the first
     * letter of each word. For instance, "RARE_CARDS" would become "Rare
     * Cards".
     * @param input string to be converted
     * @return reader friendly variant of constant name
     */
    private static String getReadableConstant(String input)
    {
        char[] name = input.toCharArray();
        
        boolean isStartOfWord = true;
        for (int i = 0; i < name.length; ++i)
        {
            char chr = name[i];
            if (chr == '_') name[i] = ' ';
            else if (isStartOfWord) name[i] = Character.toUpperCase(chr);
            else name[i] = Character.toLowerCase(chr);
            isStartOfWord = chr == '_';
        }
        
        return new String(name);
    }
    
    /**
     * Keybinding chooser with customized appearance and functionality for the
     * SIP Communicator.
     */
    private class SIPChooser
        extends BindingChooser
    {
        private static final long serialVersionUID = 0;
        
        // Provides mapping of UI labels to internal action names
        private HashMap <String, String> actionLabels =
                new HashMap <String, String>();
        
        // Calls focus to the form so keyboard events are received
        protected void onClick(MouseEvent event, BindingEntry entry,
                BindingEntry.Field field)
        {
            super.onClick(event, entry, field);
            KeybindingsConfigForm.this.requestFocus();
        }
        
        public boolean putBinding(BindingEntry newEntry, int index)
        {
            // Converts to I18N strings for UI
            String actionInternal = newEntry.getAction();
            String actionLabel = Resources.getString(actionInternal);
            this.actionLabels.put(actionLabel, actionInternal);
            newEntry.setAction(actionLabel);
            
            // Overwrites the default entry layout to stretch alt shortcut field
            newEntry.removeAll();
            newEntry.setLayout(new BorderLayout());
            
            JPanel left = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
            left.add(newEntry.getField(BindingEntry.Field.INDENT));
            left.add(newEntry.getField(BindingEntry.Field.ACTION));
            left.add(newEntry.getField(BindingEntry.Field.SHORTCUT));
            newEntry.add(left, BorderLayout.WEST);
            newEntry.add(newEntry.getField(BindingEntry.Field.ALT_SHORTCUT));
            
            return super.putBinding(newEntry, index);
        }
        
        public LinkedHashMap <KeyStroke, String> getBindingMap()
        {
            // Translates I18N strings back to internal action labels
            LinkedHashMap <KeyStroke, String> bindings =
                    new LinkedHashMap <KeyStroke, String>();
            for (BindingEntry entry : super.getBindings())
            {
                bindings.put(entry.getShortcut(), this.actionLabels.get(entry
                        .getAction()));
            }
            
            return bindings;
        }
        
        public ArrayList <BindingWrapper> getWrappedBindings() {
            // Translates I18N strings back to internal action labels
            ArrayList <BindingWrapper> bindings = new ArrayList <BindingWrapper> ();
            for (BindingEntry entry : super.getBindings())
            {
                bindings.add(new BindingWrapper(this.actionLabels.get(entry
                        .getAction()), entry.getShortcut(), entry.getAlternateShortcut()));
            }
            return bindings;
        }
    }
    
    public int getIndex()
    {
        return -1;
    }
}