/*
 * SIP Communicator, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package net.java.sip.communicator.impl.gui.keybindings;

import java.util.*;

import javax.swing.KeyStroke;

import org.osgi.framework.*;

import net.java.sip.communicator.impl.gui.GuiActivator;
import net.java.sip.communicator.service.configuration.ConfigurationService;
import net.java.sip.communicator.service.gui.keybindings.*;
import net.java.sip.communicator.service.resources.ResourceManagementService;
import net.java.sip.communicator.util.Logger;

/**
 * Service that concerns keybinding mappings used by various parts of the UI.
 * Persistence is handled as follows when started:
 * <ol>
 * <li>Load default bindings</li>
 * <li>Attempt to load custom bindings and resolve any duplicates</li>
 * <li>If merged bindings differ from the custom bindings then this attempts to
 * save the merged version</li>
 * </ol>
 * Custom bindings attempt to be written again whenever they're changed if the
 * service is running.
 * @author Damian Johnson
 */
public class KeybindingsServiceImpl
    implements KeybindingsService, Observer
{
    private static final Logger logger =
            Logger.getLogger(KeybindingsServiceImpl.class);
    
    private static final String BASE_RESOURCE_PATH =
            "net.java.sip.communicator.impl.gui.keybindings.";
    
    // Fields used in defaults.properties
    private static final String INDEX_FIELD = "index";
    private static final String SHORTCUT_FIELD = "shortcut";
    private static final String ALT_SHORTCUT_FIELD = "shortcutAlt";
    
    // hack to distinguish alternate custom bindings
    private static final String ALT_SHORTCUT_SUFFIX = "-alt";
    
    // Flag indicating if service is running
    private boolean isRunning = false;
    
    // Stores custom bindings
    private ConfigurationService configService; // stores custom bindings
    
    // Loaded keybinding mappings, maps to null if defaults failed to be loaded
    private final HashMap <KeybindingSet.Category, KeybindingSetImpl> bindings =
            new HashMap <KeybindingSet.Category, KeybindingSetImpl>();
    
    /**
     * Starts the KeybindingService, for each keybinding category retrieving the
     * default bindings then overwrites them with any custom bindings that can
     * be retrieved. This writes the merged copy back if it differs from the
     * custom bindings. This is a no-op if the service has already been started.
     * @param bc the currently valid OSGI bundle context.
     */
    public synchronized void start(BundleContext bc)
    {
        if (this.isRunning) return;
        
        // readies configuration service for loading custom bindings
        ServiceReference configServiceReference =
                bc.getServiceReference(ConfigurationService.class.getName());
        this.configService =
                (ConfigurationService) bc.getService(configServiceReference);
        try
        {
            this.configService.reloadConfiguration();
        }
        catch (Exception exc)
        {
            // silently fail - custom bindings will in turn fail to load
        }
        
        for (KeybindingSet.Category category : KeybindingSet.Category.values())
        {
            // Retrieves default bindings
            ResourceManagementService resourceService =
                    GuiActivator.getResources();
            Iterator resourceItr = resourceService.getCurrentSettings();
            
            // temporary container for loaded bindings
            ArrayList <OrderedBinding> defaultBindings =
                    new ArrayList <OrderedBinding>();
            
            // processed actions (ie, in in defaultsTmp)
            ArrayList <String> includedActions = new ArrayList <String>();
            
            while (resourceItr.hasNext())
            {
                try
                {
                    String resourceKey = (String) resourceItr.next();
                    
                    // checks if default is associated with this service
                    if (resourceKey.startsWith(BASE_RESOURCE_PATH))
                    {
                        // gets category enumerated by KeybindingSet
                        String bindingCategory =
                                resourceKey.substring(BASE_RESOURCE_PATH
                                        .length());
                        bindingCategory =
                                bindingCategory.substring(0, bindingCategory
                                        .indexOf('.'));
                        
                        // gets action associated with binding
                        String bindingAction =
                                resourceKey.substring(BASE_RESOURCE_PATH
                                        .length()
                                        + bindingCategory.length() + 1);
                        bindingAction =
                                bindingAction.substring(0, bindingAction
                                        .indexOf('.'));
                        
                        // checks if it belongs to this category and hasn't been
                        // included yet
                        if (category.getResource().equals(bindingCategory)
                                && !includedActions.contains(bindingAction))
                        {
                            int index =
                                    Integer
                                            .parseInt(resourceService
                                                    .getSettingsString(BASE_RESOURCE_PATH
                                                            + bindingCategory
                                                            + "."
                                                            + bindingAction
                                                            + "." + INDEX_FIELD));
                            KeyStroke bindingShortcut =
                                    KeyStroke
                                            .getKeyStroke(resourceService
                                                    .getSettingsString(BASE_RESOURCE_PATH
                                                            + bindingCategory
                                                            + "."
                                                            + bindingAction
                                                            + "."
                                                            + SHORTCUT_FIELD));
                            
                            // alternate shortcut (optional)
                            KeyStroke bindingAltShortcut = null;
                            String altShortcutField =
                                    BASE_RESOURCE_PATH + bindingCategory + "."
                                            + bindingAction + "."
                                            + ALT_SHORTCUT_FIELD;
                            if (resourceService
                                    .isSettingsStringAvailable(altShortcutField))
                            {
                                bindingAltShortcut =
                                        KeyStroke
                                                .getKeyStroke(resourceService
                                                        .getSettingsString(altShortcutField));
                            }
                            
                            defaultBindings.add(new OrderedBinding(
                                    bindingAction, bindingShortcut,
                                    bindingAltShortcut, index));
                            includedActions.add(bindingAction);
                        }
                    }
                }
                catch (Exception exc)
                {
                    String msg =
                            "Unable to parse default parameter for keybinding defaults";
                    logger.warn(msg, exc);
                }
                
            }
            
            Collections.sort(defaultBindings);
            
            // Attempts to retrieve custom bindings
            ArrayList <BindingWrapper> customBindings =
                    new ArrayList <BindingWrapper>();
            String setResourceName =
                    BASE_RESOURCE_PATH + category.getResource();
            List <String> customNames =
                    this.configService.getPropertyNamesByPrefix(
                            setResourceName, false);
            if (!customNames.isEmpty())
            {
                // fills with standard bindings
                for (String resourceName : customNames)
                {
                    String action =
                            resourceName.substring(resourceName
                                    .lastIndexOf('.') + 1);
                    if (action.endsWith(ALT_SHORTCUT_SUFFIX)) continue;
                    
                    KeyStroke shortcut =
                            KeyStroke.getKeyStroke((String) this.configService
                                    .getProperty(resourceName));
                    customBindings.add(new BindingWrapper(action, shortcut,
                            null));
                }
                
                // fills in alternate shortcuts
                for (String resourceName : customNames)
                {
                    String action =
                            resourceName.substring(resourceName
                                    .lastIndexOf('.') + 1);
                    if (action.endsWith(ALT_SHORTCUT_SUFFIX))
                    {
                        // removes suffix
                        action =
                                action.substring(0, action.length()
                                        - ALT_SHORTCUT_SUFFIX.length());
                        KeyStroke shortcut =
                                KeyStroke
                                        .getKeyStroke((String) this.configService
                                                .getProperty(resourceName));
                        
                        // adds to appropriate action
                        for (BindingWrapper wrapper : customBindings)
                        {
                            if (wrapper.action.equals(action)) wrapper.altShortcut =
                                    shortcut;
                        }
                        
                    }
                }
            }
            
            // Merges custom bindings
            ArrayList <BindingWrapper> merged =
                    new ArrayList <BindingWrapper>();
            if (!customBindings.isEmpty())
            {
                ArrayList <BindingWrapper> customTmp =
                        new ArrayList <BindingWrapper>(customBindings);
                
                for (BindingWrapper wrapper : defaultBindings)
                {
                    BindingWrapper associatedCustom = null;
                    for (BindingWrapper customWrapper : customTmp)
                    {
                        if (wrapper.action.equals(customWrapper.action))
                        {
                            associatedCustom = customWrapper;
                            break;
                        }
                    }
                    
                    if (associatedCustom == null)
                    {
                        merged.add(wrapper);
                    }
                    else
                    {
                        customTmp.remove(associatedCustom);
                        merged.add(associatedCustom);
                    }
                }
            }
            else
            {
                merged = new ArrayList <BindingWrapper>(defaultBindings);
            }
            
            // Writes merged result
            if (!merged.equals(customBindings))
            {
                String baseResourceName =
                        BASE_RESOURCE_PATH + category.getResource();
                for (BindingWrapper wrapper : merged)
                {
                    this.configService.setProperty(baseResourceName + "."
                            + wrapper.action, wrapper.shortcut.toString());
                    if (wrapper.altShortcut != null)
                    {
                        this.configService.setProperty(baseResourceName + "."
                                + wrapper.action + ALT_SHORTCUT_SUFFIX,
                                wrapper.altShortcut.toString());
                    }
                }
            }
            
            KeybindingSetImpl newSet =
                    new KeybindingSetImpl(merged,
                            new ArrayList <BindingWrapper>(defaultBindings),
                            category);
            this.bindings.put(category, newSet);
            newSet.addObserver(this);
        }
        
        this.isRunning = true;
    }
    
    /**
     * Invalidates references to custom bindings, preventing further writes.
     */
    public synchronized void stop()
    {
        if (!this.isRunning) return;
        this.bindings.clear();
        this.isRunning = false;
    }
    
    /**
     * Provides the bindings associated with a given category. This may be null
     * if the default bindings failed to be loaded.
     * @param category segment of the UI for which bindings should be retrieved
     * @return mappings of keystrokes to the string representation of their
     *         actions
     * @throws UnsupportedOperationException if the service isn't running
     */
    public synchronized KeybindingSet getBindings(
            KeybindingSet.Category category)
    {
        if (!this.isRunning) throw new UnsupportedOperationException();
        
        // Started services should have all categories
        assert this.bindings.containsKey(category);
        return this.bindings.get(category);
    }
    
    // wrapper for ordered binding data
    private class OrderedBinding
        extends BindingWrapper
        implements Comparable <OrderedBinding>
    {
        Integer index;
        
        OrderedBinding(String action, KeyStroke shortcut,
                KeyStroke altShortcut, int index)
        {
            super(action, shortcut, altShortcut);
            this.index = new Integer(index);
        }
        
        public int compareTo(OrderedBinding binding)
        {
            return this.index.compareTo(binding.index);
        }
    }
    
    // Listens for changes in binding sets so changes can be written
    public void update(Observable obs, Object arg)
    {
        if (obs instanceof KeybindingSetImpl)
        {
            KeybindingSetImpl changedSet = (KeybindingSetImpl) obs;
            
            // Writes merged result
            synchronized (this)
            {
                String baseResourceName =
                        BASE_RESOURCE_PATH
                                + changedSet.getCategory().getResource();
                
                for (BindingWrapper wrapper : changedSet.getWrappedBindings())
                {
                    this.configService.setProperty(baseResourceName + "."
                            + wrapper.action, wrapper.shortcut.toString());
                    
                    String altValue = wrapper.altShortcut == null ? "" : wrapper.altShortcut.toString();
                    this.configService.setProperty(baseResourceName + "."
                            + wrapper.action + ALT_SHORTCUT_SUFFIX,
                            altValue);
                }
            }
        }
    }
}