/*
 * KeybindingsUtil, open source library for keybinding customization and management.
 * Project Page: http://www.atagar.com/keybindingUtil/
 * Available under the Apache v2 (http://www.apache.org/licenses/LICENSE-2.0.html)
 * or LGPL v2 (http://www.gnu.org/licenses/old-licenses/gpl-2.0.html)
 */
package net.java.sip.communicator.impl.gui.keybindings.chooser;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;

import net.java.sip.communicator.impl.gui.keybindings.chooser.util.ComponentTools;
import net.java.sip.communicator.impl.gui.keybindings.chooser.util.StringTools;

/**
 * Implementation of the BindingPanel that provides configuring functionality
 * for the keystroke component of key bindings. Methods provide a means of
 * producing predefined, sweeping changes in the display. This defaults to a
 * light blue color scheme with an index indent style.<br>
 * Though display elements are still accessible, manual changes are not
 * particularly recommended unless automated changes to the appearance (the
 * indentation style and color scheme) are disabled since they may be
 * unexpectedly reverted or clash any alterations made.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version September 1, 2007
 */
public class BindingChooser
    extends BindingPanel
{
    private static final long serialVersionUID = 0;
    private IndentStyle indentStyle = IndentStyle.INDEX;
    private boolean isShortcutEditable = true; // Determines if shortcut fields
    // can be selected
    private BindingEntry selectedEntry = null; // None selected when null
    private boolean isAlternateSelected = false;
    private String selectedText = "Press shortcut...";
    
    // Color scheme applied to contents
    private boolean isColorSchemeEnabled = true;
    private Color selectedEntryBackground;
    private Color evenEntryBackground;
    private Color oddEntryBackground;
    private Color entryForeground;
    private Color selectedEntryForeground;
    
    {
        setColorScheme(new Color(110, 110, 255), true); // Sets default color
        // scheme
    }
    
    /**
     * Displays a dialog allowing the user to redefine the keystroke component
     * of key bindings. The top has light blue labels describing the fields and
     * the bottom provides an 'OK' and 'Cancel' option. This uses the default
     * color scheme and indent style. If no entries are selected then the enter
     * key is equivalent to pressing 'OK' and escape is the same as 'Cancel'.
     * @param parent frame to which to apply modal property and center within
     *            (centers within screen if null)
     * @param bindings initial mapping of keystrokes to their actions
     * @return redefined mapping of keystrokes to their actions, null if cancel
     *         is pressed
     */
    public static LinkedHashMap <KeyStroke, String> showDialog(
            Component parent, Map <KeyStroke, String> bindings)
    {
        BindingChooser display = new BindingChooser();
        display.putAllBindings(bindings);
        return showDialog(parent, display, "Key Bindings", true, display
                .makeAdaptor(), FrameColorScheme.HUE_DERIVED);
    }
    
    /**
     * Displays a dialog allowing the user to redefine the keystroke component
     * of key bindings. The bottom provides an 'OK' and 'Cancel' option. If no
     * entries are selected then the enter key is equivalent to pressing 'OK'
     * and escape is the same as 'Cancel'. Label and button backgrounds try to
     * match color scheme if set.<br>
     * Including focusable elements in the display will prevent user input from
     * setting the selected shortcut field. Also note that labels use the
     * default entry size and should be omitted if using content with custom
     * dimensions.
     * @param parent frame to which to apply modal property and center within
     *            (centers within screen if null)
     * @param display body of the display, containing current bindings and
     *            appearance properties
     * @param dialogTitle title of the displayed dialog
     * @param showLabels if true the top has labels describing the fields,
     *            otherwise they are omitted
     * @param adaptor adaptor used to provide configuring functionality
     * @param frameColoring method of determining coloring of labels and buttons
     * @return redefined mapping of keystrokes to their actions, null if cancel
     *         is pressed
     */
    public static LinkedHashMap <KeyStroke, String> showDialog(
            Component parent, final BindingChooser display, String dialogTitle,
            boolean showLabels, BindingAdaptor adaptor,
            FrameColorScheme frameColoring)
    {
        final JDialog dialog = ComponentTools.makeDialog(parent, dialogTitle);
        dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        dialog.setResizable(false);
        
        if (showLabels) dialog.add(display.makeLabels(frameColoring),
                BorderLayout.NORTH);
        dialog.add(display);
        
        // Bottom controls
        JPanel controlSection = new JPanel(new GridLayout(1, 0));
        Color reference = frameColoring.getBackgroundBasis(display);
        
        // HACK: Uses button's name as a mutable value to determine if pressed
        final String PRESSED_STATE = "pressed";
        final JButton okButton =
                ComponentTools.makeThemedButton("OK", reference);
        okButton.setName("not " + PRESSED_STATE);
        okButton.setPreferredSize(new Dimension(
                okButton.getPreferredSize().width, 25));
        okButton.setFocusable(false);
        okButton.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                okButton.setName(PRESSED_STATE);
                dialog.dispose();
            }
        });
        controlSection.add(okButton);
        
        final JButton cancelButton =
                ComponentTools.makeThemedButton("Cancel", reference);
        cancelButton.setPreferredSize(new Dimension(cancelButton
                .getPreferredSize().width, 25));
        cancelButton.setFocusable(false);
        cancelButton.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                dialog.dispose();
            }
        });
        controlSection.add(cancelButton);
        dialog.add(controlSection, BorderLayout.SOUTH);
        
        // Adds listener that closes dialog when pressing 'enter' or 'escape'
        dialog.addKeyListener(new KeyAdapter()
        {
            public void keyPressed(KeyEvent event)
            {
                if (display.selectedEntry == null)
                {
                    if (event.getKeyCode() == KeyEvent.VK_ENTER)
                    {
                        okButton.doClick();
                    }
                    else if (event.getKeyCode() == KeyEvent.VK_ESCAPE)
                    {
                        cancelButton.doClick();
                    }
                }
            }
        });
        dialog.addKeyListener(adaptor); // Listener for shortcut field input
        dialog.pack();
        
        // Centers dialog
        if (parent != null) ComponentTools.center(dialog, parent);
        else ComponentTools.center(dialog);
        
        dialog.setVisible(true);
        if (okButton.getName().equals(PRESSED_STATE)) return display
                .getBindingMap();
        else return null;
    }
    
    /*-
     * This is called upon:
     * Component reordering (inherited functionality from BindingPanel)
     * Visual changes to the entry
     * Component validation
     */
    protected void onUpdate(int index, BindingEntry entry, boolean isNew)
    {
        applyColorScheme(entry, index);
        this.indentStyle.apply(entry, index);
    }
    
    protected void onClick(MouseEvent event, BindingEntry entry,
            BindingEntry.Field field)
    {
        // Selects shortcut fields when they're clicked
        if (field == BindingEntry.Field.SHORTCUT
                || field == BindingEntry.Field.ALT_SHORTCUT)
        {
            // Deselects if already selected
            if (entry.equals(this.selectedEntry)) setSelected(null, false);
            else
            {
                setSelected(entry, field == BindingEntry.Field.ALT_SHORTCUT);
            }
        }
    }
    
    /**
     * Sets if the shortcut fields of entries can be selected to provide editing
     * functionality or not. If false, any selected entry is deselected.
     * @param editable if true shortcut fields may be selected to have their
     *            values changed, otherwise user input and calls to the
     *            setSelected method are ignored
     */
    public void setEditable(boolean editable)
    {
        if (!editable && this.selectedEntry != null)
        {
            setSelected(null, false); // Deselects current selection
        }
        
        this.isShortcutEditable = editable;
    }
    
    /**
     * Sets if the display should manage its coloring scheme or not.
     * @param enabled if true then display changes may trigger the recoloring of
     *            components according to the current color scheme, otherwise
     *            automated recoloring is disabled
     */
    public void setColorSchemeEnabled(boolean enabled)
    {
        this.isColorSchemeEnabled = enabled;
    }
    
    /**
     * Sets foreground and background of components to a theme centered around a
     * given hue. This effects components in the following ways:<br>
     * Selected Field (Background)- Given color.<br>
     * Selected Field (Foreground)- Black unless given color is very dark.<br>
     * Even Entries (Background)- White if lightened is true, black otherwise.<br>
     * Odd Entries (Background)- Lightened hue of color if lightened is true,
     * darkened otherwise.<br>
     * All Entries (Foreground)- Black if lightened is true, white otherwise.<br>
     * If color schemes are enabled then this will immediately apply the changes
     * to the display.
     * @param color hue considered in setting component colors
     * @param lightened if true backgrounds are bent toward a lighter theme,
     *            otherwise they are darker
     */
    public void setColorScheme(Color color, boolean lightened)
    {
        int r = 85 * color.getRed() / 255;
        int g = 85 * color.getGreen() / 255;
        int b = 85 * color.getBlue() / 255;
        
        Color oddBackground;
        if (lightened) oddBackground = new Color(170 + r, 170 + g, 170 + b);
        else oddBackground = new Color(r, g, b);
        
        Color evenBackground = lightened ? Color.WHITE : Color.BLACK;
        Color selectedBackground = color;
        Color foreground = lightened ? Color.BLACK : Color.WHITE;
        
        int rgbSum = color.getRed() + color.getGreen() + color.getBlue();
        Color selectedForeground = rgbSum > 256 ? Color.BLACK : Color.WHITE;
        setColorScheme(selectedBackground, evenBackground, oddBackground,
                foreground, selectedForeground);
    }
    
    /**
     * Manually sets each attribute of color scheme. If color schemes are
     * enabled then this will immediately apply the changes to the display. This
     * will prompt an onUpdate on all entries.
     * @param selectedBackground shortcut field's background color when awaiting
     *            user input
     * @param evenBackground background color of entries with an even index
     *            (starting with zero)
     * @param oddBackground background color of entries with an odd index
     *            (starting with zero)
     * @param foreground text color of entries
     * @param selectedForeground text color of entries awaiting input
     */
    public void setColorScheme(Color selectedBackground, Color evenBackground,
            Color oddBackground, Color foreground, Color selectedForeground)
    {
        this.selectedEntryBackground = selectedBackground;
        this.evenEntryBackground = evenBackground;
        this.oddEntryBackground = oddBackground;
        this.entryForeground = foreground;
        this.selectedEntryForeground = selectedForeground;
        
        ArrayList <BindingEntry> bindings = getBindings();
        for (int i = 0; i < bindings.size(); ++i)
        {
            onUpdate(i, bindings.get(i), false);
        }
    }
    
    /**
     * Provides the indent style used by the chooser.
     * @return type of content in the indent field
     */
    public IndentStyle getIndentStyle()
    {
        return this.indentStyle;
    }
    
    /**
     * Sets content display in the indent field of entries. This will prompt an
     * onUpdate on all entries unless setting the style to NONE.
     * @param style type of content displayed in entry's indent field
     */
    public void setIndentStyle(IndentStyle style)
    {
        this.indentStyle = style;
        
        if (style == IndentStyle.NONE) return;
        ArrayList <BindingEntry> bindings = getBindings();
        for (int i = 0; i < bindings.size(); ++i)
        {
            onUpdate(i, bindings.get(i), false);
        }
    }
    
    /**
     * Sets the message of the selected shortcut field when awaiting user input.
     * By default this is "Press shortcut...".
     * @param message prompt for user input
     */
    public void setSelectedText(String message)
    {
        if (this.selectedEntry != null)
        {
            this.selectedEntry.getField(BindingEntry.Field.SHORTCUT).setText(
                    message);
        }
        
        this.selectedText = message;
    }
    
    /**
     * Returns if a binding is currently awaiting input or not.
     * @return true if a binding is awaiting input, false otherwise
     */
    public boolean isBindingSelected()
    {
        return this.selectedEntry != null;
    }
    
    /**
     * Returns if the selected component is the alternate (false if none is
     * selected).
     * @return true if alternate binding is awaiting input, false otherwise
     */
    public boolean isAlternateSelected()
    {
        if (!isBindingSelected()) return false;
        return this.isAlternateSelected;
    }
    
    /**
     * Provides the currently selected entry if awaiting input.
     * @return entry currently awaiting input, if one exists
     */
    public BindingEntry getSelected()
    {
        return this.selectedEntry;
    }
    
    /**
     * Sets the shortcut field of an entry to prompt user input. The next call
     * to doInput sets set its shortcut field and deselects the entry. Any other
     * currently selected entry is deselected. If null, then this simply reverts
     * any selections (leaving no entry selected). The onUpdate method is called
     * whenever an entry is either selected or deselected.
     * @param entry binding entry awaiting input for its shortcut field
     * @throws IllegalArgumentException if entry is not contained in chooser
     */
    public void setSelected(BindingEntry entry, boolean isAlternate)
    {
        if (!this.isShortcutEditable) return; // Selection can't be changed
        
        // Entry is already selected
        if (entry != null && entry.equals(this.selectedEntry)
                && this.isAlternateSelected == isAlternate) return;
        
        if (entry != null && !getBindings().contains(entry))
        {
            throw new IllegalArgumentException(
                    "BindingEntry not contained in display.");
        }
        
        BindingEntry previousSelection = this.selectedEntry;
        boolean previousIsAlternate = this.isAlternateSelected;
        this.selectedEntry = entry;
        this.isAlternateSelected = isAlternate;
        
        // Reverts previously selected field's attributes
        if (previousSelection != null)
        {
            onUpdate(getBindingIndex(previousSelection), previousSelection,
                    false);
            if (previousIsAlternate)
            {
                previousSelection.setAlternateShortcut(previousSelection
                        .getAlternateShortcut()); // Reverts
                // text
            }
            else
            {
                previousSelection.setShortcut(previousSelection.getShortcut()); // Reverts
                // text
            }
        }
        
        // Sets the new selection
        if (this.selectedEntry != null)
        {
            onUpdate(getBindingIndex(this.selectedEntry), this.selectedEntry,
                    false);
            
            if (this.isAlternateSelected)
            {
                this.selectedEntry.getField(BindingEntry.Field.ALT_SHORTCUT)
                        .setText(" " + this.selectedText);
            }
            else
            {
                this.selectedEntry.getField(BindingEntry.Field.SHORTCUT)
                        .setText(" " + this.selectedText);
            }
        }
    }
    
    /**
     * Provides a key adaptor that can provide editing functionality for the
     * selected entry.
     * @return binding adaptor configured to this chooser
     */
    public BindingAdaptor makeAdaptor()
    {
        return new BindingAdaptor(this);
    }
    
    /**
     * Provides the labels naming the fields. These are based on the settings
     * when constructed and aren't updated when the display changes. Labels use
     * the default entry dimensions.
     * @param colorScheme method of setting background gradient
     * @return labels used in dialog
     */
    public BindingEntry makeLabels(FrameColorScheme colorScheme)
    {
        // Uses a BindingEntry to provide default dimensions
        // Applies theme background to the BindingEntry and the foreground and
        // borders to the fields
        final Color reference = colorScheme.getBackgroundBasis(this);
        BindingEntry labels = new BindingEntry(null, null, "")
        {
            private static final long serialVersionUID = 0;
            
            public void paintComponent(Graphics g)
            {
                Graphics2D g2 = (Graphics2D) g;
                g2.setPaint(ComponentTools.applyThemedScheme(null, reference,
                        false));
                g2.fillRect(0, 0, getWidth(), getHeight()); // Draws gradient
                // background
                super.paintComponent(g);
            }
        };
        labels.setOpaque(false);
        
        for (BindingEntry.Field field : BindingEntry.Field.values())
        {
            JLabel fieldLabel = labels.getField(field);
            ComponentTools.applyThemedScheme(fieldLabel, reference, false);
            
            if (field == BindingEntry.Field.INDENT)
            {
                // Removes indent field if omitted from the rest of the display.
                fieldLabel.setVisible(indentStyle != IndentStyle.EMPTY);
            }
            else if (field == BindingEntry.Field.ACTION)
            {
                fieldLabel.setText(" Action:");
            }
            else if (field == BindingEntry.Field.SHORTCUT)
            {
                fieldLabel.setText(" Shortcut:");
            }
            else if (field == BindingEntry.Field.ALT_SHORTCUT)
            {
                fieldLabel.setText(" Alternate:");
            }
            else
            {
                // BindingEntry.Field has changed and this should be updated
                // accordingly
                assert false : BindingChooser.class.getName()
                        + " doesn't recognize the '" + field + "' field.";
            }
        }
        return labels;
    }
    
    // Sets the colors of the entry according to the current color scheme if
    // enabled
    protected void applyColorScheme(BindingEntry entry, int index)
    {
        if (this.isColorSchemeEnabled)
        {
            for (BindingEntry.Field field : BindingEntry.Field.values())
            {
                JLabel fieldLabel = entry.getField(field);
                fieldLabel.setOpaque(true);
                
                Color background =
                        index % 2 == 0 ? this.evenEntryBackground
                                : this.oddEntryBackground;
                fieldLabel.setBackground(background);
                fieldLabel.setForeground(this.entryForeground);
            }
            
            if (entry.equals(this.selectedEntry))
            {
                JLabel shortcutField =
                        this.isAlternateSelected ? entry
                                .getField(BindingEntry.Field.ALT_SHORTCUT)
                                : entry.getField(BindingEntry.Field.SHORTCUT);
                shortcutField.setBackground(this.selectedEntryBackground);
                shortcutField.setForeground(this.selectedEntryForeground);
            }
        }
    }
    
    /**
     * Emulates keyboard input, setting the selected entry's shortcut if an
     * entry's currently awaiting input.
     * @param input keystroke input for selected entry
     */
    void doInput(KeyStroke input)
    {
        if (isBindingSelected())
        {
            if (this.isAlternateSelected) this.selectedEntry
                    .setAlternateShortcut(input);
            else this.selectedEntry.setShortcut(input);
            
            // TYPE indent can change according to the shortcut
            // this.indentStyle.apply(this.selectedEntry,
            // getBindingIndex(this.selectedEntry));
            setSelected(null, false); // Deselects shortcut field
        }
    }
    
    @Override
    public void validate()
    {
        super.validate();
        
        ArrayList <BindingEntry> bindings = getBindings();
        for (int i = 0; i < bindings.size(); ++i)
        {
            onUpdate(i, bindings.get(i), false);
        }
    }
    
    /**
     * Supported appearances of the indent field, which includes:<br>
     * NONE- No actions are taken to change the indent field's appearance.<br>
     * EMPTY- Indent field is set to be invisible (effectively removing it from
     * the display).<br>
     * SPACER- Blank field that occupies its currently set dimensions.<br>
     * TYPE- Displays Unicode arrows according to the shortcut's event type
     * (down for KEY_PRESSED, up for KEY_RELEASED, bidirectional for KEY_TYPED,
     * and an 'X' if disabled).<br>
     * INDEX- Displays the field's index from the top (starting with one).
     */
    public static enum IndentStyle
    {
        NONE, EMPTY, SPACER, TYPE, INDEX;
        
        /**
         * Returns the enum representation of a string. This is case sensitive.
         * @param str toString representation of this enum
         * @return enum associated with a string
         * @throws IllegalArgumentException if argument is not represented by
         *             this enum.
         */
        public static IndentStyle fromString(String str)
        {
            for (IndentStyle type : IndentStyle.values())
            {
                if (str.equals(type.toString())) return type;
            }
            throw new IllegalArgumentException();
        }
        
        // Applies this style to the indent field of an entry. This uses a zero
        // based index.
        private void apply(BindingEntry entry, int index)
        {
            if (this == NONE) return;
            JLabel indentField = entry.getField(BindingEntry.Field.INDENT);
            indentField.setVisible(this != EMPTY);
            indentField.setIcon(null);
            
            String fieldText = "";
            if (this == TYPE)
            {
                if (entry.getShortcut() == BindingEntry.DISABLED)
                {
                    fieldText = "  X";
                }
                else
                {
                    int type = entry.getShortcut().getKeyEventType();
                    if (type == KeyEvent.KEY_PRESSED) fieldText = "  \u2193";
                    else if (type == KeyEvent.KEY_RELEASED) fieldText =
                            "  \u2191";
                    else if (type == KeyEvent.KEY_TYPED) fieldText = "  \u2195";
                    else
                    {
                        // Should be unreachable according to the AWTKeyStroke
                        // class
                        assert false : "Unrecognized key type: " + type;
                        fieldText = "";
                    }
                }
            }
            else if (this == INDEX)
            {
                fieldText = " " + (index + 1) + ".";
            }
            indentField.setText(fieldText);
        }
        
        @Override
        public String toString()
        {
            if (this == TYPE) return "Event Type";
            return StringTools.getReadableConstant(this.name());
        }
    }
    
    /**
     * Coloring used for the labels and buttons of constructed dialogs. If the
     * color scheme is not set, then colors based on it will default to being
     * based on gray. This influences the content as follows:<br>
     * GRAY- Sets backgrounds to gradients based on gray.<br>
     * GRAY_DERIVED- Sets backgrounds to gray scale gradients based on the color
     * scheme.<br>
     * HUE_DERIVED- Sets backgrounds to colored gradients based on the color
     * scheme.
     */
    public static enum FrameColorScheme
    {
        GRAY, GRAY_DERIVED, HUE_DERIVED;
        
        /**
         * Returns the enum representation of a string. This is case sensitive.
         * @param str toString representation of this enum
         * @return enum associated with a string
         * @throws IllegalArgumentException if argument is not represented by
         *             this enum.
         */
        public static FrameColorScheme fromString(String str)
        {
            for (FrameColorScheme type : FrameColorScheme.values())
            {
                if (str.equals(type.toString())) return type;
            }
            throw new IllegalArgumentException();
        }
        
        // Provides the color on which the dialog's label and button backgrounds
        // are based
        public Color getBackgroundBasis(BindingChooser display)
        {
            if (this == GRAY || !display.isColorSchemeEnabled) return Color.GRAY;
            else if (this == GRAY_DERIVED)
            {
                // Gets the average rgb of the content's background colors
                int rgbAvg = display.evenEntryBackground.getRed();
                rgbAvg += display.evenEntryBackground.getGreen();
                rgbAvg += display.evenEntryBackground.getBlue();
                rgbAvg += display.oddEntryBackground.getRed();
                rgbAvg += display.oddEntryBackground.getGreen();
                rgbAvg += display.oddEntryBackground.getBlue();
                rgbAvg = rgbAvg / 6;
                
                return new Color(rgbAvg, rgbAvg, rgbAvg).darker();
            }
            else if (this == HUE_DERIVED)
            {
                // Gets average rgb components of content's background colors
                Color even = display.evenEntryBackground;
                Color odd = display.oddEntryBackground;
                
                int r = (even.getRed() + odd.getRed()) / 2;
                int g = (even.getGreen() + odd.getGreen()) / 2;
                int b = (even.getBlue() + odd.getBlue()) / 2;
                return new Color(r, g, b).darker();
            }
            else
            {
                assert false : "Unrecognized FrameColorScheme type: " + this;
                return Color.GRAY;
            }
        }
        
        @Override
        public String toString()
        {
            return StringTools.getReadableConstant(this.name());
        }
    }
}
