/*
 * KeybindingsUtil, open source library for keybinding customization and management.
 * Project Page: http://www.atagar.com/keybindingUtil/
 * Available under the Apache v2 (http://www.apache.org/licenses/LICENSE-2.0.html)
 * or LGPL v2 (http://www.gnu.org/licenses/old-licenses/gpl-2.0.html)
 */
package net.java.sip.communicator.impl.gui.keybindings.chooser;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

/**
 * Display element for a single key binding.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version August 7, 2007
 */
public class BindingEntry
    extends JPanel
{
    /**
     * Keystroke representing an unmapped entry.
     */
    public static final KeyStroke DISABLED = null;
    
    // Display dimensions
    private static final int LINE_HEIGHT = 25;
    private static final int INDENT_WIDTH = 25;
    private static final int ACTION_WIDTH = 150;
    private static final int SHORTCUT_WIDTH = 150;
    private static final int ALT_SHORTCUT_WIDTH = 150;
    
    private static final long serialVersionUID = 0;
    private JLabel indentField = new JLabel();
    private JLabel actionField = new JLabel();
    private JLabel shortcutField = new JLabel();
    private JLabel altShortcutField = new JLabel();
    private KeyStroke shortcut;
    private KeyStroke altShortcut;
    
    {
        this.indentField.setPreferredSize(new Dimension(INDENT_WIDTH,
                LINE_HEIGHT));
        this.indentField.setForeground(Color.BLACK);
        
        this.actionField.setPreferredSize(new Dimension(ACTION_WIDTH,
                LINE_HEIGHT));
        this.actionField.setForeground(Color.BLACK);
        
        this.shortcutField.setPreferredSize(new Dimension(SHORTCUT_WIDTH,
                LINE_HEIGHT));
        this.shortcutField.setForeground(Color.BLACK);
        
        this.altShortcutField.setPreferredSize(new Dimension(ALT_SHORTCUT_WIDTH,
                LINE_HEIGHT));
        this.altShortcutField.setForeground(Color.BLACK);
    }
    
    public BindingEntry(KeyStroke shortcut, KeyStroke altShortcut, String action)
    {
        setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
        setAction(action);
        setShortcut(shortcut);
        setAlternateShortcut(altShortcut);
        
        add(this.indentField);
        add(this.actionField);
        add(this.shortcutField);
        add(this.altShortcutField);
    }
    
    public KeyStroke getShortcut()
    {
        return this.shortcut;
    }
    
    public void setShortcut(KeyStroke shortcut)
    {
        setShortcut(shortcut, true);
    }
    
    public KeyStroke getAlternateShortcut()
    {
        return this.altShortcut;
    }
    
    public void setAlternateShortcut(KeyStroke shortcut)
    {
        setShortcut(shortcut, false);
    }
    
    private void setShortcut(KeyStroke shortcut, boolean isPrimary)
    {
        if (isPrimary) this.shortcut = shortcut;
        else this.altShortcut = shortcut;
        
        JLabel field = isPrimary ? this.shortcutField : this.altShortcutField;
        
        // Sets shortcut label to representation of the keystroke
        if (shortcut == DISABLED)
        {
            if (isPrimary) field.setText(" Disabled");
            else field.setText(" ");
        }
        else
        {
            StringBuffer buffer = new StringBuffer();
            
            if (shortcut.getKeyEventType() == KeyEvent.KEY_TYPED)
            {
                buffer.append(shortcut.getKeyChar());
            }
            else
            {
                int keycode = shortcut.getKeyCode();
                int modifiers = shortcut.getModifiers();
                
                // Indicates modifiers of the keystroke
                boolean shiftMask = (modifiers & InputEvent.SHIFT_MASK) != 0;
                boolean ctrlMask = (modifiers & InputEvent.CTRL_MASK) != 0;
                boolean metaMask = (modifiers & InputEvent.META_MASK) != 0;
                boolean altMask = (modifiers & InputEvent.ALT_MASK) != 0;
                if (shiftMask && keycode != KeyEvent.VK_SHIFT) buffer
                        .append("Shift + ");
                if (ctrlMask && keycode != KeyEvent.VK_CONTROL) buffer
                        .append("Ctrl + ");
                if (metaMask && keycode != KeyEvent.VK_META) buffer
                        .append("Meta + ");
                if (altMask && keycode != KeyEvent.VK_ALT) buffer
                        .append("Alt + ");
                
                buffer.append(KeyEvent.getKeyText(keycode));
            }
            
            field.setText(" " + buffer.toString());
        }
    }
    
    public String getAction()
    {
        return this.actionField.getText().substring(1);
    }
    
    public void setAction(String action)
    {
        this.actionField.setText(" " + action);
    }
    
    public boolean isPrimaryDisabled()
    {
        return this.shortcut == DISABLED;
    }
    
    public boolean isAlternateDisabled()
    {
        return this.altShortcut == DISABLED;
    }
    
    public boolean isAltDisabled()
    {
        return this.altShortcut == DISABLED;
    }
    
    /**
     * Provides the label associated with a field.
     * @param field element of display to be returned
     * @return label associated with field
     */
    public JLabel getField(Field field)
    {
        if (field == Field.INDENT) return this.indentField;
        else if (field == Field.ACTION) return this.actionField;
        else if (field == Field.SHORTCUT) return this.shortcutField;
        else if (field == Field.ALT_SHORTCUT) return this.altShortcutField;
        else
        {
            assert false : "Unrecognized field: " + field;
            return null;
        }
    }
    
    /**
     * Elements of the display (ordered left to right):<br>
     * Indent- Leading index label or icon.<br>
     * Action- String component of mapping.<br>
     * Shortcut- Keystroke component of mapping.
     * Alt_Shortcut- Alternate keystroke mapping.
     */
    public enum Field
    {
        INDENT, ACTION, SHORTCUT, ALT_SHORTCUT;
    }
    
    /**
     * Provides the string representation of this mapping. The exact details of
     * the representation are unspecified and subject to change but the
     * following format can be considered to be typical:<br>
     * "BindingEntry (" + Shortcut + " \u2192 " + Action + ")"
     * @return string representation of entry
     */
    @Override
    public String toString()
    {
        StringBuilder builder = new StringBuilder();
        builder.append("BindingEntry (");
        
        if (isPrimaryDisabled()) builder.append("Disabled");
        else builder.append(getShortcut());
        builder.append(" & ");
        if (isAlternateDisabled()) builder.append("Disabled");
        else builder.append(getAlternateShortcut());
        
        builder.append(" \u2192 "); // Adds Unicode arrow pointing right
        builder.append(getAction());
        builder.append(")");
        return builder.toString();
    }
    
    /**
     * Checks if argument is an instance of this class with the same shortcut
     * and associated action. It does not compare aspects of the display
     * elements.
     * @param obj element with which to be compared
     * @return true if argument is an instance of this class with matching
     *         shortcut and action, false otherwise
     */
    @Override
    public boolean equals(Object obj)
    {
        if (obj == this) return true;
        else if (!(obj instanceof BindingEntry)) return false;
        
        BindingEntry entry = (BindingEntry) obj;
        boolean equals = true;
        
        String action = this.getAction();
        if (action == null) equals &= entry.getAction() == null;
        else equals &= action.equals(entry.getAction());
        
        KeyStroke shortcut = this.getShortcut();
        if (shortcut == null) equals &= entry.getShortcut() == null;
        else equals &= shortcut.equals(entry.getShortcut());
        
        return equals;
    }
    
    @Override
    public int hashCode()
    {
        int hash = 17;
        hash = 37 * hash + (getAction() == null ? 0 : getAction().hashCode());
        hash =
                37
                        * hash
                        + (getShortcut() == null ? 0 : getShortcut().hashCode());
        return hash;
    }
}
