/*
 * SIP Communicator, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package net.java.sip.communicator.plugin.spellchecker;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.PixelGrabber;
import java.awt.image.WritableRaster;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import javax.swing.*;

import net.java.sip.communicator.service.contactlist.*;
import net.java.sip.communicator.service.gui.*;
import net.java.sip.communicator.service.gui.Container;
import net.java.sip.communicator.util.*;

/**
 * Combo box providing a listing of all available locales with corresponding
 * country flags. Selecting a new field causes that locale's dictionary to be
 * downloaded, if not available. The spell checker then use the selected
 * language for further checking.
 * @author Damian Johnson
 */
public class LanguageSelectionField
    implements PluginComponent
{
    private static final HashMap <SpellChecker, LanguageSelectionField> CLASS_INSTANCES =
            new HashMap <SpellChecker, LanguageSelectionField>();
    
    // parallel maps containing cached instances of country flags
    private static final HashMap <Parameters.Locale, ImageIcon> AVAILABLE_FLAGS =
            new HashMap <Parameters.Locale, ImageIcon>();
    private static final HashMap <Parameters.Locale, ImageIcon> UNAVAILABLE_FLAGS =
            new HashMap <Parameters.Locale, ImageIcon>();
    
    private static final Logger logger =
            Logger.getLogger(LanguageSelectionField.class);
    private static final ImageIcon BLANK_FLAG_ICON =
            Resources.getImage("blankFlag");
    private static final int LANGUAGE_SELECTION_FIELD_WIDTH = 225;
    private final ListCellRenderer languageSelectionRenderer;
    private final HashMap <Parameters.Locale, Boolean> localeAvailabilityCache =
            new HashMap <Parameters.Locale, Boolean>();
    private final SpellChecker spellChecker;
    private final ArrayList <JComboBox> fieldInstances =
            new ArrayList <JComboBox>();
    private JComboBox pluginInstance = null; // instance for settings
    
    /**
     * Provides instance of this class associated with a spell checker. If ones
     * already been created then this instance is used.
     * @param checker spell checker field is to be associated with
     * @return spell checker locale selection field
     */
    public synchronized static LanguageSelectionField makeSelectionField(
            SpellChecker checker)
    {
        // singleton constructor to ensure only one combo box is associated with
        // each checker
        if (CLASS_INSTANCES.containsKey(checker)) return CLASS_INSTANCES
                .get(checker);
        else
        {
            LanguageSelectionField instance =
                    new LanguageSelectionField(checker);
            CLASS_INSTANCES.put(checker, instance);
            return instance;
        }
    }
    
    private LanguageSelectionField(SpellChecker checker)
    {
        this.spellChecker = checker;
        this.languageSelectionRenderer = new ListCellRenderer()
        {
            public Component getListCellRendererComponent(JList list,
                    Object value, int index, boolean isSelected,
                    boolean cellHasFocus)
            {
                Parameters.Locale locale = (Parameters.Locale) value;
                
                if (!localeAvailabilityCache.containsKey(locale))
                {
                    localeAvailabilityCache.put(locale, spellChecker
                            .isLocaleAvailable(locale));
                }
                
                ImageIcon flagIcon =
                        getLocaleIcon(locale, localeAvailabilityCache
                                .get(locale));
                
                JLabel label = new JLabel();
                label.setText(locale.getLabel());
                label.setIcon(flagIcon);
                return label;
            }
        };
    }
    
    public JComboBox makeField()
    {
        JComboBox field = new JComboBox();
        field.setRenderer(this.languageSelectionRenderer);
        for (Parameters.Locale locale : Parameters.getLocales())
        {
            field.addItem(locale);
        }
        field.setSelectedItem(this.spellChecker.getLocale());
        
        field.addActionListener(new ActionListener()
        {
            private boolean skipFiring = false;
            private Parameters.Locale previousLocale = spellChecker.getLocale();
            
            public void actionPerformed(ActionEvent event)
            {
                // prevents potential infinite loop during errors
                if (this.skipFiring) return;
                final JComboBox source = (JComboBox) event.getSource();
                final Parameters.Locale locale =
                        (Parameters.Locale) source.getSelectedItem();
                if (locale.equals(this.previousLocale)) return;
                this.previousLocale = locale;
                
                source.setEnabled(false);
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                        try
                        {
                            spellChecker.setLocale(locale);
                            source.setEnabled(true);
                            localeAvailabilityCache.put(locale, true);
                            
                            // resets all other instances
                            for (JComboBox field : fieldInstances)
                            {
                                if (!field.equals(source))
                                {
                                    field.setSelectedItem(locale);
                                }
                            }
                        }
                        catch (Exception exc)
                        {
                            logger.warn("Unable to retrieve dictionary for "
                                    + locale, exc);
                            
                            // warns that it didn't work
                            PopupDialog dialog =
                                    spellChecker.getUIService()
                                            .getPopupDialog();
                            String message =
                                    Resources.getString("switchDictError");
                            if (exc instanceof IOException)
                            {
                                message =
                                        Resources
                                                .getString("retrieveDictError")
                                                + ":\n" + locale.getDictUrl();
                            }
                            else if (exc instanceof IllegalArgumentException)
                            {
                                message =
                                        Resources.getString("processDictError");
                            }
                            dialog.showMessagePopupDialog(message, Resources
                                    .getString("switchDictErrorTitle"),
                                    PopupDialog.WARNING_MESSAGE);
                            
                            // reverts selection
                            skipFiring = true;
                            source.setSelectedItem(spellChecker.getLocale());
                            skipFiring = false;
                            source.setEnabled(true);
                        }
                    }
                });
            }
        });
        
        int height = field.getPreferredSize().height;
        field.setPreferredSize(new Dimension(LANGUAGE_SELECTION_FIELD_WIDTH,
                height));
        
        this.fieldInstances.add(field);
        return field;
    }
    
    public ArrayList <JMenuItem> getAvailableLocaleListing()
    {
        ArrayList <JMenuItem> availableLocales = new ArrayList <JMenuItem> ();
        
        for (final Parameters.Locale locale : Parameters.getLocales())
        {
            if (!this.localeAvailabilityCache.containsKey(locale))
            {
                localeAvailabilityCache.put(locale, spellChecker
                        .isLocaleAvailable(locale));
            }
            
            // skips if locale isn't available
            if (!this.localeAvailabilityCache.get(locale)) continue;
            
            JMenuItem localeItem = new JMenuItem(locale.getLabel(), getLocaleIcon(locale, true));
            
            // sets selected locale to be bold
            Font labelFont = localeItem.getFont();
            if (spellChecker.getLocale().equals(locale)) 
                localeItem.setFont(new Font(labelFont.getFontName(), Font.BOLD, labelFont.getSize()));
            
            // switches selected locale
            localeItem.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    SwingUtilities.invokeLater(new Runnable()
                    {
                        public void run()
                        {
                            try
                            {
                                spellChecker.setLocale(locale);
                                
                                // resets selected field of combo boxes
                                for (JComboBox field : fieldInstances)
                                {
                                    field.setSelectedItem(locale);
                                }
                            }
                            catch (Exception exc)
                            {
                                logger.warn("Unable to retrieve dictionary for "
                                        + locale, exc);
                            }
                        }
                    });
                }
            });
            
            availableLocales.add(localeItem);
        }
        
        return availableLocales;
    }
    
    /**
     * Clears any cached data used by the field so it reflects the current state
     * of its associated spell checker.
     */
    public void revalidate()
    {
        this.localeAvailabilityCache.clear();
    }
    
    public Object getComponent()
    {
        if (this.pluginInstance == null) this.pluginInstance = makeField();
        
        return this.pluginInstance;
    }
    
    public String getConstraints()
    {
        return Container.RIGHT;
    }
    
    public Container getContainer()
    {
        return Container.CONTAINER_CHAT_TOOL_BAR;
    }
    
    public String getName()
    {
        return "Spell Checker Toggle";
    }
    
    public int getPositionIndex()
    {
        return -1;
    }
    
    public boolean isNativeComponent()
    {
        return false;
    }
    
    public void setCurrentContact(MetaContact metaContact)
    {

    }
    
    public void setCurrentContactGroup(MetaContactGroup metaGroup)
    {

    }
    
    private static ImageIcon getLocaleIcon(Parameters.Locale locale,
            boolean isAvailable)
    {
        if (isAvailable && AVAILABLE_FLAGS.containsKey(locale)) return AVAILABLE_FLAGS
                .get(locale);
        else if (!isAvailable && UNAVAILABLE_FLAGS.containsKey(locale)) return UNAVAILABLE_FLAGS
                .get(locale);
        else
        {
            // load resource
            ImageIcon localeFlag;
            
            try
            {
                int commaIndex = locale.getIsoCode().indexOf(",");
                String countryCode =
                        locale.getIsoCode().substring(commaIndex + 1);
                localeFlag = Resources.getFlagImage(countryCode);
                
                BufferedImage flagBuffer = copy(localeFlag.getImage());
                setFaded(flagBuffer);
                ImageIcon unavailableLocaleFlag = new ImageIcon(flagBuffer);
                
                AVAILABLE_FLAGS.put(locale, localeFlag);
                UNAVAILABLE_FLAGS.put(locale, unavailableLocaleFlag);
                return isAvailable ? localeFlag : unavailableLocaleFlag;
            }
            catch (IOException exc)
            {
                AVAILABLE_FLAGS.put(locale, BLANK_FLAG_ICON);
                UNAVAILABLE_FLAGS.put(locale, BLANK_FLAG_ICON);
                return BLANK_FLAG_ICON;
            }
        }
    }
    
    /**
     * Creates a deep copy of an image.
     * @param image picture to be processed
     * @return copy of the image
     */
    private static BufferedImage copy(Image image)
    {
        int width = image.getWidth(null);
        int height = image.getHeight(null);
        
        BufferedImage copy;
        try
        {
            PixelGrabber pg = new PixelGrabber(image, 0, 0, 1, 1, false);
            pg.grabPixels();
            ColorModel cm = pg.getColorModel();
            
            WritableRaster raster =
                    cm.createCompatibleWritableRaster(width, height);
            boolean isRasterPremultiplied = cm.isAlphaPremultiplied();
            copy = new BufferedImage(cm, raster, isRasterPremultiplied, null);
        }
        catch (InterruptedException e)
        {
            copy =
                    new BufferedImage(width, height,
                            BufferedImage.TYPE_INT_ARGB);
        }
        
        Graphics2D g2 = copy.createGraphics();
        g2.setComposite(AlphaComposite.Src); // Preserves color of
        // transparent pixels
        g2.drawImage(image, 0, 0, null);
        g2.dispose();
        return copy;
    }
    
    /**
     * Removes all color from an image and makes it partly translucent. Original
     * grayscale method written by Marty Stepp.
     * @param image picture to be processed
     */
    private static void setFaded(BufferedImage image)
    {
        int width = image.getWidth();
        int height = image.getHeight();
        
        for (int row = 0; row < width; ++row)
        {
            for (int col = 0; col < height; ++col)
            {
                int c = image.getRGB(row, col);
                
                int r =
                        (((c >> 16) & 0xff) + ((c >> 8) & 0xff) + (c & 0xff)) / 3;
                
                int newRgb = (0xff << 24) | (r << 16) | (r << 8) | r;
                newRgb &= (1 << 24) - 1; // Blanks alpha value
                newRgb |= 128 << 24; // Resets it to the alpha of 128
                image.setRGB(row, col, newRgb);
            }
        }
    }
}
