/*
 * SIP Communicator, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package net.java.sip.communicator.plugin.spellchecker;

import java.net.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import net.java.sip.communicator.service.resources.ResourceManagementService;
import net.java.sip.communicator.util.Logger;

/**
 * Information provided via the spellchecer's default parameters.
 * @author Damian Johnson
 */
class Parameters
{
    private static final Logger logger = Logger.getLogger(Parameters.class);
    private static final String PLUGIN_PREFIX = "net.java.sip.communicator.plugin.spellchecker.";
    private static final String DEFAULT_PREFIX = "default.";
    private static final String LOCALE_PREFIX = "locale.";
    private static final String LOCALE_LABEL_FIELD = "label";
    private static final String LOCALE_DICT_FIELD = "dict";
    
    private static final HashMap <Default, String> DEFAULTS =
            new HashMap <Default, String>();
    private static final ArrayList <Locale> LOCALES = new ArrayList <Locale>();
    
    static void init(ResourceManagementService service) 
    {
        // parses defaults
        DEFAULTS.clear();
        for (String key : service.getSettingsKeys(PLUGIN_PREFIX + DEFAULT_PREFIX))
        {
            try {
                DEFAULTS.put(Default.fromString(key), service.getSettingsString(key));
            } catch (IllegalArgumentException exc) {
                logger.error("Unrecognized spell checker default: " + key);
            }
        }
        
        // parse locales
        LOCALES.clear();
        int prefixSize = PLUGIN_PREFIX.length() + LOCALE_PREFIX.length();
        HashSet <String> processedIsoCodes = new HashSet <String> ();
        for (String key : service.getSettingsKeys(PLUGIN_PREFIX + LOCALE_PREFIX))
        {
            String isoCode = key.substring(prefixSize, prefixSize + 5);
            if (processedIsoCodes.contains(isoCode)) continue;
            String label = service.getSettingsString(PLUGIN_PREFIX + LOCALE_PREFIX + isoCode + "." + LOCALE_LABEL_FIELD);
            String dict = service.getSettingsString(PLUGIN_PREFIX + LOCALE_PREFIX + isoCode + "." + LOCALE_DICT_FIELD);
            processedIsoCodes.add(isoCode);
            
            if (label == null || dict == null) {
                logger.error("Unable to find all parameters for spell checker locale: " + isoCode);
                continue;
            }
            
            URL dictUrl;
            try {
                dictUrl = new URL(dict);
            } catch (MalformedURLException exc) {
                logger.error("Unable to parse dictionary url for spell checker locale: " + isoCode);
                continue;
            }
            
            LOCALES.add(new Locale(label, isoCode, dictUrl));
        }
    }
    
    /**
     * Provides the value of a particular default field, null if undefined.
     * @param field default field to retrieve
     * @return value corresponding to default field
     */
    public static String getDefault(Default field)
    {
        return DEFAULTS.get(field);
    }
    
    /**
     * Provides locale with a given iso code. Null if undefined.
     * @param isoCode iso code of locale to be retrieved
     * @return locale with corresponding iso code
     */
    public static Locale getLocale(String isoCode)
    {
        for (Locale locale : LOCALES)
        {
            if (locale.getIsoCode().equals(isoCode)) return locale;
        }
        
        return null;
    }
    
    /**
     * Provides locales in which dictionary resources are available.
     * @return locations with dictionary resources
     */
    public static ArrayList <Locale> getLocales()
    {
        return new ArrayList <Locale>(LOCALES);
    }
    
    /**
     * Locale with an available dictionary resource.
     */
    public static class Locale
    {
        private final String label;
        private final String isoCode;
        private final URL dictLocation;
        
        private Locale(String label, String isoCode, URL dictLocation)
        {
            this.label = label;
            this.isoCode = isoCode;
            this.dictLocation = dictLocation;
        }
        
        /**
         * Provides user readable name of language.
         * @return name of language presented to user
         */
        public String getLabel()
        {
            return this.label;
        }
        
        /**
         * Provides ISO code as defined by:<br />
         * http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
         * @return iso code
         */
        public String getIsoCode()
        {
            return this.isoCode;
        }
        
        /**
         * Provides the url where the dictionary resource can be found for this
         * language.
         * @return url of dictionary resource
         */
        public URL getDictUrl()
        {
            return this.dictLocation;
        }
        
        @Override
        public String toString()
        {
            return this.label + " (" + this.isoCode + ")";
        }
    }
    
    /**
     * Default attribute that may be defined in the parameters xml.
     */
    public enum Default
    {
        LOCALE("locale");
        
        private String tag;
        
        Default(String tag)
        {
            this.tag = tag;
        }
        
        /**
         * Returns the enum representation of a string. This is case sensitive.
         * @param str toString representation of a default field
         * @return default field associated with a string
         * @throws IllegalArgumentException if argument is not represented by a
         *             default field.
         */
        public static Default fromString(String str)
        {
            for (Default field : Default.values())
            {
                if (str.equals(field.toString()))
                {
                    return field;
                }
            }
            throw new IllegalArgumentException();
        }
        
        @Override
        public String toString()
        {
            return this.tag;
        }
    }
}
