/*
 * SIP Communicator, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package net.java.sip.communicator.plugin.spellchecker;

import java.awt.image.*;
import java.io.*;
import java.util.*;

import javax.imageio.*;
import javax.swing.*;

import net.java.sip.communicator.util.*;

/**
 * The <tt>Resources</tt> class manages the access to the internationalization
 * properties files and the image resources used in this plugin.
 * @author Damian Johnson
 */
public class Resources
{
    private static Logger log = Logger.getLogger(Resources.class);
    
    /**
     * The name of the resource, where paths to images used in this bundle are
     * stored.
     */
    private static final String IMAGE_RESOURCE_NAME =
            "net.java.sip.communicator.plugin.spellchecker.resources";
    
    /**
     * Location of flag resources.
     */
    private static final String FLAG_PATH =
            "resources/images/plugin/spellchecker/flags/";
    
    // Available images
    public static final String PLUGIN_ICON = "pluginIcon";
    public static final String ADD_WORD_ICON = "addWordIcon";
    public static final String PERSONAL_DICTIONARY = "personalDictionary";
    public static final String WORD_INCLUDE = "wordInclude";
    public static final String WORD_EXCLUDE = "wordExclude";
    
    /**
     * The image resource bundle.
     */
    private static final ResourceBundle IMAGE_RESOURCE_BUNDLE =
            ResourceBundle.getBundle(IMAGE_RESOURCE_NAME);
    
    /**
     * Returns an internationalized string corresponding to the given key.
     * @param key The key of the string.
     * @return An internationalized string corresponding to the given key.
     */
    public static String getString(String key)
    {
        return SpellCheckerActivator.getResources().getI18NString(key);
    }
    
    /**
     * Loads an image from a given image identifier.
     * @param imageID The identifier of the image.
     * @return The image for the given identifier.
     */
    public static ImageIcon getImage(String imageID)
    {
        BufferedImage image = null;
        
        String path = IMAGE_RESOURCE_BUNDLE.getString(imageID);
        try
        {
            image =
                    ImageIO.read(Resources.class.getClassLoader()
                            .getResourceAsStream(path));
        }
        catch (IOException e)
        {
            log.error("Failed to load image:" + path, e);
        }
        
        return new ImageIcon(image);
    }
    
    /**
     * Loads a flag image from a given image identifier.
     * @param resource iso code for flag to be retrieved.
     * @return icon reflecting iso code
     * @throws IOException if no such resource is available
     */
    public static ImageIcon getFlagImage(String resource) throws IOException
    {
        String path = FLAG_PATH + resource + ".png";
        InputStream input =
                Resources.class.getClassLoader().getResourceAsStream(path);
        if (input == null) throw new IOException("Unable to retrieve: " + path);
        BufferedImage image = ImageIO.read(input);
        return new ImageIcon(image);
    }
    
    /**
     * Loads an image from a given image identifier.
     * @param imageID The identifier of the image.
     * @return The image for the given identifier.
     */
    public static byte[] getImageInBytes(String imageID)
    {
        byte[] image = new byte[100000];
        
        String path = IMAGE_RESOURCE_BUNDLE.getString(imageID);
        
        try
        {
            Resources.class.getClassLoader().getResourceAsStream(path).read(
                    image);
        }
        catch (IOException e)
        {
            log.error("Failed to load image:" + path, e);
        }
        
        return image;
    }
}
