/*
 * SIP Communicator, the OpenSource Java VoIP and Instant Messaging client.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package net.java.sip.communicator.plugin.spellchecker;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.List;

import javax.swing.*;
import javax.swing.text.*;

/**
 * Dialog providing the ability to edit a list of words (adding and marking for
 * removal). This prevents duplicate entries.
 * @author Damian Johnson
 */
class WordlistCustomizationDialog
    extends JDialog
{
    private static final long serialVersionUID = 0;
    private static final int HGAP = 25; // extra width beyond content
    private static final String TITLE;
    private static final ImageIcon PERSONAL_DICTIONARY_ICON;
    private static final ImageIcon INCLUDE_WORD_ICON; // check mark
    private static final ImageIcon EXCLUDE_WORD_ICON; // red 'X'
    private static final Dimension WORD_PANEL_SIZE = new Dimension(100, 400);
    
    // Background colors to accent selections
    private static final Color INCLUDE_WORD_BG_LIGHT = new Color(220, 255, 220);
    private static final Color INCLUDE_WORD_BG_DARK = new Color(175, 255, 175);
    private static final Color EXCLUDE_WORD_BG_LIGHT = new Color(255, 235, 235);
    private static final Color EXCLUDE_WORD_BG_DARK = new Color(255, 200, 200);
    
    // Mapping between listed words and if they should be included
    private LinkedHashMap <String, Boolean> wordList =
            new LinkedHashMap <String, Boolean>();
    private JPanel wordPanel = new JPanel(new GridLayout(0, 1));
    private JScrollPane wordScroller;
    private boolean applyPressed = false;
    
    static
    {
        TITLE = Resources.getString("personalDictionary");
        PERSONAL_DICTIONARY_ICON =
                Resources.getImage(Resources.PERSONAL_DICTIONARY);
        INCLUDE_WORD_ICON = Resources.getImage(Resources.WORD_INCLUDE);
        EXCLUDE_WORD_ICON = Resources.getImage(Resources.WORD_EXCLUDE);
    }
    
    /**
     * Shows a dialog allowing the customization of a list of words.
     * @param owner owning frame to which modal property is applied
     * @param initialWords initial listing of words, duplicate entries are
     *            stripped
     * @return resulting list of words, null if unchanged or not applied
     */
    public static ArrayList <String> showDialog(Frame owner,
            List <String> initialWords)
    {
        // strips duplicates
        Set <String> tmp = new LinkedHashSet <String>(initialWords);
        ArrayList <String> origWords = new ArrayList <String>(tmp);
        Collections.sort(origWords); // alphabetizes
        
        WordlistCustomizationDialog dialog =
                new WordlistCustomizationDialog(owner, origWords);
        if (owner != null) center(dialog, owner); // centers within owner
        dialog.setVisible(true);
        
        if (dialog.applyPressed)
        {
            // Constructs listing of all selected words
            boolean isSame = true; // selection differs from original words
            ArrayList <String> newWords = new ArrayList <String>();
            for (String possibleWord : dialog.wordList.keySet())
            {
                if (dialog.wordList.get(possibleWord))
                {
                    newWords.add(possibleWord);
                    if (!origWords.remove(possibleWord)) isSame = false;
                }
            }
            isSame &= origWords.isEmpty();
            
            if (isSame) return null;
            else return newWords;
        }
        else return null;
    }
    
    /**
     * Constructs a new modal dialog that allows the customization of a word
     * list.
     * @param owner frame to which modal property is applied
     * @param initialWords initial listing of words
     */
    private WordlistCustomizationDialog(Frame owner, List <String> initialWords)
    {
        super(owner, true);
        setLayout(new BorderLayout());
        setTitle(TITLE);
        
        // Wrapper to constrain height to preferred size
        JPanel wrapper = new JPanel(new BorderLayout());
        wrapper.add(this.wordPanel, BorderLayout.NORTH);
        this.wordScroller = new JScrollPane(wrapper);
        this.wordScroller.setPreferredSize(WORD_PANEL_SIZE);
        this.wordScroller
                .setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        add(this.wordScroller);
        
        // Adds all initial words
        for (String initialWord : initialWords)
        {
            addWord(initialWord);
        }
        
        // Top banner
        JLabel banner =
                new JLabel(TITLE, PERSONAL_DICTIONARY_ICON,
                        SwingConstants.CENTER);
        banner.setForeground(Color.BLACK);
        banner.setFont(new Font("Serif", Font.ITALIC, 45));
        add(banner, BorderLayout.NORTH);
        
        // Bottom controls
        JPanel bottom = new JPanel(new BorderLayout());
        
        // Controls for adding words
        JPanel lowerLeft = new JPanel(new FlowLayout(FlowLayout.LEFT));
        JTextField addWordField = new JTextField(10);
        addWordField.addActionListener(new AddWordListener(addWordField));
        lowerLeft.add(addWordField);
        
        JButton addWordButton = new JButton(Resources.getString("addWord"));
        addWordButton.addActionListener(new AddWordListener(addWordField));
        lowerLeft.add(addWordButton);
        bottom.add(lowerLeft, BorderLayout.WEST);
        
        // Apply and Cancel buttons
        JPanel lowerRight = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        JButton ok = new JButton(Resources.getString("ok"));
        ok.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                applyPressed = true;
                dispose();
            }
        });
        lowerRight.add(ok);
        
        JButton cancel = new JButton(Resources.getString("cancel"));
        cancel.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent event)
            {
                applyPressed = false;
                dispose();
            }
        });
        lowerRight.add(cancel);
        bottom.add(lowerRight, BorderLayout.EAST);
        
        add(bottom, BorderLayout.SOUTH);
        pack();
        
        // Adds extra space to keep dialog from being cramped
        Dimension d = getSize();
        d.width += HGAP;
        setSize(d);
    }
    
    // Adds a word to both the view and the model.
    private void addWord(final String word)
    {
        if (!this.wordList.containsKey(word))
        {
            this.wordList.put(word, true);
            
            JCheckBox wordSelection = new JCheckBox(word, false);
            wordSelection.setForeground(Color.BLACK);
            wordSelection.setSelectedIcon(INCLUDE_WORD_ICON);
            wordSelection.setIcon(EXCLUDE_WORD_ICON);
            wordSelection.setFocusable(false); // prevents focus box
            wordSelection.addActionListener(new ActionListener()
            {
                private final boolean isLight = wordList.size() % 2 == 1;
                
                public void actionPerformed(ActionEvent event)
                {
                    JCheckBox source = (JCheckBox) event.getSource();
                    boolean isSelected = source.isSelected();
                    wordList.put(word, isSelected);
                    Color background;
                    if (isSelected)
                    {
                        background =
                                isLight ? INCLUDE_WORD_BG_LIGHT
                                        : INCLUDE_WORD_BG_DARK;
                    }
                    else
                    {
                        background =
                                isLight ? EXCLUDE_WORD_BG_LIGHT
                                        : EXCLUDE_WORD_BG_DARK;
                    }
                    source.setBackground(background);
                }
            });
            
            // sets selected and initializes appearance
            wordSelection.doClick();
            
            this.wordPanel.add(wordSelection);
            
            // shows entry and updates scroll bar
            this.wordScroller.validate();
        }
    }
    
    // Adds word stored in text component and warns if it's already included
    private class AddWordListener
        implements ActionListener
    {
        private JTextComponent textField;
        
        private AddWordListener(JTextComponent textField)
        {
            this.textField = textField;
        }
        
        public void actionPerformed(ActionEvent event)
        {
            String word = this.textField.getText();
            if (wordList.containsKey(word))
            {
                // displays warning dialog if list already contains the word
                String message =
                        Resources.getString("duplicateWarning") + " \"" + word
                                + "\"";
                JOptionPane.showMessageDialog(WordlistCustomizationDialog.this,
                        message);
            }
            else
            {
                addWord(word);
                this.textField.setText(""); // clears text field
            }
        }
    }
    
    /**
     * Moves a window to be concentric with another.
     * @param mover window to be centered
     * @param target to be centered within
     */
    public static void center(Window mover, Component target)
    {
        Dimension moverSize = mover.getSize();
        Dimension targetSize = target.getSize();
        int x = (targetSize.width - moverSize.width) / 2;
        x += target.getLocation().x;
        int y = (targetSize.height - moverSize.height) / 2;
        y += target.getLocation().y;
        mover.setLocation(x, y);
    }
}
